/*
 *
 *  Copyright (C) 1999 Peter Amstutz
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of *the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA
 */

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import com.sun.j3d.utils.applet.MainFrame;
import com.sun.j3d.utils.universe.*;
import com.sun.j3d.utils.geometry.*;
import javax.media.j3d.*;
import javax.vecmath.*;
import java.util.*;

public class MovementBehavior extends Behavior {
    MovementBehavior(TransformGroup rot, TransformGroup pos, TransformGroup av) {
        cameraRot=rot;
        cameraPos=pos;
        avatar=av;
    }

    protected TransformGroup cameraRot;
    protected TransformGroup cameraPos;
    protected TransformGroup avatar;

    protected Alpha movealph;
    protected Alpha rotalph;
    protected double lastmovealpha=0.0;
    protected double lastrotalpha=0.0;
    protected Vector3d direction;
    protected Vector3d orientation;
    protected long movetime=0;
    protected long rottime=0;
    protected boolean moving;

    public boolean isMoving() { return moving; }
    public void setMoving(boolean m) { moving=m; }

    public void initialize() {
        wakeupOn(new WakeupOnElapsedFrames(0));
    }

    public void startMovement(Vector3d dir, long dur) {
        direction=dir;
        movetime=dur;
        lastmovealpha=0.0;
        movealph=new Alpha(1, dur);
    }

    public void startRotation(Vector3d angles, long dur) {
        orientation=angles;
        rottime=dur;
        lastrotalpha=0.0;
        rotalph=new Alpha(1, dur);
    }

    protected Matrix3d rot=new Matrix3d();
    protected Vector3d pos=new Vector3d();
    protected Transform3D tmp=new Transform3D();
    protected Transform3D cpos=new Transform3D();
    protected Transform3D crotx=new Transform3D();
    protected Transform3D croty=new Transform3D();
    protected Transform3D crotz=new Transform3D();

    public void processStimulus(Enumeration criteria) {
        if(movealph!=null) {
            double alphadelta=movealph.value() - lastmovealpha;
            lastmovealpha=movealph.value();

            cameraPos.getTransform(cpos);
            cpos.get(pos);
            cameraRot.getTransform(tmp);
            tmp.get(rot);
            rot.invert();

            tmp.setRotation(rot);
            Vector3d result=new Vector3d(direction.x * alphadelta,
                                         direction.y * alphadelta,
                                         direction.z * alphadelta);
            tmp.transform(result);
            result.add(pos);
            cpos.setTranslation(result);
            cameraPos.setTransform(cpos);

            result.negate();
            cpos.setTranslation(result);
            avatar.setTransform(cpos);

            if(movealph.finished()) {
                if(moving) {
                    lastmovealpha=0.0;
                    movealph=new Alpha(1, movetime);
                } else movealph=null;
            }
        }
        wakeupOn(new WakeupOnElapsedFrames(0));
    }
}
