/*
 *
 *  Copyright (C) 1999 Peter Amstutz
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of *the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA
 */

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import com.sun.j3d.utils.applet.MainFrame;
import com.sun.j3d.utils.universe.*;
import com.sun.j3d.utils.geometry.*;
import javax.media.j3d.*;
import javax.vecmath.*;
import java.util.*;

// avg high 1.75 m

class Terrain implements MouseListener, MouseMotionListener {
    TransformGroup cameraRot=new TransformGroup();
    TransformGroup cameraPos=new TransformGroup();
    TransformGroup objectTransformGroup=new TransformGroup();
    BranchGroup objectBranchGroup=new BranchGroup();
    BranchGroup masterBranchGroup=new BranchGroup();
    SimpleUniverse universe;
    MovementBehavior movement, gravity;
    int oldMouseX, oldMouseY;
    boolean mouse1down=false;
    boolean mouse2down=false;
    boolean mouse3down=false;
    Random random=new Random();
    JFrame mainframe;
    com.sun.j3d.utils.geometry.Box avatar;
    boolean usegravity=false;

    Terrain() {
        mainframe=new JFrame("Terrain test");
        mainframe.setSize(640, 480);
        mainframe.getContentPane().setLayout(new BorderLayout());

        mainframe.addWindowListener(new WindowAdapter()
                                  {
                                      public void windowClosing(WindowEvent e) {
                                          System.exit(0);
                                      }
                                  });

        Canvas3D c3d=new Canvas3D(GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice().getDefaultConfiguration());
        c3d.addMouseListener(this);
        c3d.addMouseMotionListener(this);
        mainframe.getContentPane().add(c3d, "Center");

        // set up camera
        cameraPos.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
        cameraPos.setCapability(Group.ALLOW_CHILDREN_EXTEND);
        cameraPos.setCapability(Group.ALLOW_CHILDREN_WRITE);
        cameraPos.setCapability(Group.ALLOW_CHILDREN_READ);
        cameraPos.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
        cameraPos.setCapability(Geometry.ALLOW_INTERSECT);

        cameraRot.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
        cameraRot.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
        cameraRot.setCapability(Geometry.ALLOW_INTERSECT);
        cameraRot.addChild(cameraPos);

        objectTransformGroup.setCapability(Group.ALLOW_CHILDREN_EXTEND);
        objectTransformGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
        objectTransformGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);

        objectBranchGroup.addChild(objectTransformGroup);
        cameraPos.addChild(objectBranchGroup);

        createLights();

        TransformGroup avatargroup=new TransformGroup();

        // set up movement behavior
        movement=new MovementBehavior(cameraRot, cameraPos, avatargroup);
        movement.setSchedulingBounds(new BoundingSphere());
        cameraRot.addChild(movement);
        movement.setEnable(false);

        // set up gravity behavior
        if(usegravity) {
            gravity=new GravityBehavior(cameraRot, cameraPos, avatargroup);
            gravity.setSchedulingBounds(new BoundingSphere());
            cameraRot.addChild(gravity);
            gravity.setEnable(false);
        }

        Transform3D t3d=new Transform3D();

        // set up collision
        if(usegravity) {
        // set up avatar
            TransformGroup avatarpush=new TransformGroup();
            avatar=new com.sun.j3d.utils.geometry.Box((float).6, (float)1.75,
                                                      (float).3, new Appearance());
            avatarpush.addChild(avatar);
            t3d.setTranslation(new Vector3d(0, -.85, 0));
            avatarpush.setTransform(t3d);
            avatargroup.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
            avatargroup.addChild(avatarpush);
            objectTransformGroup.addChild(avatargroup);

            CollisionBehavior collision=new CollisionBehavior(avatar, gravity);
            collision.setSchedulingBounds(new BoundingSphere(new Point3d(0, 0, 0), 10));
            cameraRot.addChild(collision);
            collision.setEnable(true);
        }

        // set up universe
        universe=new SimpleUniverse(c3d);
        universe.getViewer().getView().setBackClipDistance(20);
        //universe.getViewingPlatform().setPlatformGeometry(pg);
        masterBranchGroup=new BranchGroup();
        masterBranchGroup.setCapability(Group.ALLOW_CHILDREN_READ);
        masterBranchGroup.setCapability(BranchGroup.ALLOW_DETACH);
        masterBranchGroup.addChild(cameraRot);
        universe.addBranchGraph(masterBranchGroup);

        createTerrain();

        t3d=new Transform3D();
        t3d.setTranslation(new Vector3d(-10, -23, -10));
        cameraPos.setTransform(t3d);

        Yangle=3.14/2;

        Transform3D crotx=new Transform3D();
        Transform3D croty=new Transform3D();
        Transform3D crotz=new Transform3D();

        crotx.rotX(Xangle);
        croty.rotY(Yangle);
        crotz.rotZ(Zangle);
        crotx.mul(croty);
        crotx.mul(crotz);
        cameraRot.setTransform(crotx);

        mainframe.show();
    }

    void createLights() {
        // set up lights
        // white light

        TransformGroup lighttransform=new TransformGroup();
        PointLight pt=new PointLight(new Color3f((float).8, (float).8, (float).8),
                                     new Point3f(0, 0, 0),
                                     new Point3f((float)1, (float).01, 0));
        pt.setInfluencingBounds(new BoundingSphere(new Point3d(0, 0, 0), 20));
        lighttransform.addChild(pt);
        Transform3D t3d=new Transform3D();
        t3d.setTranslation(new Vector3d(55, 65, 55));
        lighttransform.setTransform(t3d);
        Appearance ap=new Appearance();
        Material mat=new Material(new Color3f(0, 0, 0), new Color3f((float).6, (float).6, (float).6),
                                  new Color3f(0, 0, 0), new Color3f(0, 0, 0), 0);
        ap.setMaterial(mat);
        Sphere sph=new Sphere(5, ap);
        lighttransform.addChild(sph);
        objectBranchGroup.addChild(lighttransform);


          /*
          // red light
          TransformGroup lighttransform=new TransformGroup();
          PointLight pt=new PointLight(new Color3f((float).8, (float).1, (float).1),
          new Point3f(0, 0, 0),
          new Point3f((float)1, (float).01, 0));
          pt.setInfluencingBounds(new BoundingSphere(new Point3d(0, 0, 0), 20));
          lighttransform.addChild(pt);
          Transform3D t3d=new Transform3D();
          t3d.setTranslation(new Vector3d(35, 25, 35));
          lighttransform.setTransform(t3d);
          Appearance ap=new Appearance();
          Material mat=new Material(new Color3f(0, 0, 0), new Color3f((float).7, 0, 0),
          new Color3f(0, 0, 0), new Color3f(0, 0, 0), 0);
          ap.setMaterial(mat);
          Sphere sph=new Sphere(1, ap);
          lighttransform.addChild(sph);
          objectBranchGroup.addChild(lighttransform);

          // green light
          lighttransform=new TransformGroup();
          pt=new PointLight(new Color3f((float).1, (float).8, (float).1),
          new Point3f(0, 0, 0),
          new Point3f((float)1, (float).01, 0));
          pt.setInfluencingBounds(new BoundingSphere(new Point3d(0, 0, 0), 20));
          lighttransform.addChild(pt);
          t3d=new Transform3D();
          t3d.setTranslation(new Vector3d(95, 25, 45));
          lighttransform.setTransform(t3d);
          ap=new Appearance();
          mat=new Material(new Color3f(0, 0, 0), new Color3f(0, (float).7, 0),
          new Color3f(0, 0, 0), new Color3f(0, 0, 0), 0);
          ap.setMaterial(mat);
          sph=new Sphere(1, ap);
          lighttransform.addChild(sph);
          objectBranchGroup.addChild(lighttransform);

          // blue light
          lighttransform=new TransformGroup();
          pt=new PointLight(new Color3f((float).1, (float).1, (float).8),
          new Point3f(0, 0, 0), new Point3f((float)1, (float).01, 0));
          pt.setInfluencingBounds(new BoundingSphere(new Point3d(0, 0, 0), 20));
          lighttransform.addChild(pt);
          t3d=new Transform3D();
          t3d.setTranslation(new Vector3d(45, 25, 95));
          lighttransform.setTransform(t3d);
          ap=new Appearance();
          mat=new Material(new Color3f(0, 0, 0), new Color3f(0, 0, (float).7),
          new Color3f(0, 0, 0), new Color3f(0, 0, 0), 0);
          ap.setMaterial(mat);
          sph=new Sphere(1, ap);
          lighttransform.addChild(sph);
          objectBranchGroup.addChild(lighttransform);
        */
    }

    void createTerrain() {
        // create terrain
        double[][] buffer=new double[30][30];
        buffer[0][0]=3000;
        buffer[0][buffer[0].length-1]=3000;
        buffer[buffer.length-1][0]=3000;
        buffer[buffer.length-1][buffer[0].length-1]=3000;

        //int[] terraintable={4, 10, 100, 250, 500, 1000, 2000, 2500, 3000};
        int[] terraintable={4, 10, 50, 75, 150, 375, 750, 1500, 3000};

        tessalate(buffer, 0, 0, buffer.length-1, buffer[0].length-1, 8, terraintable);

        double[][] hights=blur(buffer);
        BranchGroup terrain=makeTerrain(hights);
        //BranchGroup terrain=makeTerrain(buffer);
        terrain.compile();
        objectTransformGroup.addChild(terrain);

        createTrees(hights);
    }

    void createTrees(double[][] hights) {
        double[][] buffer=new double[hights.length][hights[0].length];
        buffer[0][0]=3000;
        buffer[0][buffer[0].length-1]=3000;
        buffer[buffer.length-1][0]=3000;
        buffer[buffer.length-1][buffer[0].length-1]=3000;

        int[] treetable={4, 12, 12, 60, 120, 600, 900, 1800, 3000}; // adds up to about 6500

        tessalate(buffer, 0, 0, buffer.length-1, buffer[0].length-1, 8, treetable);

        TransformGroup scalegroup=new TransformGroup();
        Transform3D t3d=new Transform3D();
        TransformGroup tg;
        int x, y;
        for(y=1; y<buffer[0].length-1; y++) {
            for(x=1; x<buffer.length-1; x++) {
                int n=random.nextInt()%6500+6500;
                //System.out.println("n is "+n+" * 3000 = "+(n*3000)+" buffer is "+buffer[x][y]);
                if(n*2 < buffer[x][y]) {
                    // leaf
                    Appearance ap=new Appearance();
                    Material mat=new Material(new Color3f(0, 0, 0), new Color3f(0, 0, 0),
                                              new Color3f((float).1, (float).8, (float).3),
                                              new Color3f((float).1, (float).8, (float).3), 32);
                    ap.setMaterial(mat);
                    Cone leaf=new Cone((float).4, 300, Primitive.GENERATE_NORMALS, ap);
                    TransformGroup leafgroup=new TransformGroup();
                    t3d.setIdentity();
                    t3d.setTranslation(new Vector3d(0, 180, 0));
                    leafgroup.setTransform(t3d);
                    leafgroup.addChild(leaf);

                    // trunk
                    ap=new Appearance();
                    mat=new Material(new Color3f(0, 0, 0), new Color3f(0, 0, 0),
                                              new Color3f((float).7, (float).8, (float).1),
                                              new Color3f((float).7, (float).8, (float).1), 32);
                    ap.setMaterial(mat);
                    Cylinder trunk=new Cylinder((float).08, 80, Primitive.GENERATE_NORMALS, ap);
                    TransformGroup trunkgroup=new TransformGroup();
                    t3d.setIdentity();
                    t3d.setTranslation(new Vector3d(0, 40, 0));
                    trunkgroup.setTransform(t3d);
                    trunkgroup.addChild(trunk);

                    t3d.setIdentity();
                    t3d.setTranslation(new Vector3d(x, hights[x][y], y));

                    tg=new TransformGroup();
                    tg.addChild(leafgroup);
                    tg.addChild(trunkgroup);
                    tg.setTransform(t3d);
                    scalegroup.addChild(tg);
                }
            }
        }
        t3d=new Transform3D();
        t3d.setScale(new Vector3d(4, .015, 4));
        scalegroup.setTransform(t3d);

        BranchGroup bg=new BranchGroup();
        bg.addChild(scalegroup);
        bg.compile();
        objectTransformGroup.addChild(bg);
    }

    double[][] blur(double[][] buffer)
    {
        double[][] buffer2=new double[buffer.length][buffer[0].length];
        int x, y;

        for(y=1; y<buffer[0].length-1; y++) {
            for(x=1; x<buffer.length-1; x++) {
                buffer2[x][y]=(buffer[x-1][y-1]
                               +buffer[x][y-1]
                               +buffer[x+1][y-1]
                               +buffer[x-1][y]
                               +buffer[x][y]
                               +buffer[x+1][y]
                               +buffer[x-1][y+1]
                               +buffer[x][y+1]
                               +buffer[x+1][y+1])/9;
            }
        }
        return buffer2;
    }

    void tessalate(double[][] buffer, int x1, int y1, int x2, int y2, int depth, int[] variation) {
        int mx=x1+(x2-x1)/2;
        int my=y1+(y2-y1)/2;

        if(x1==x2 || x1+1==x2) return;

        if(buffer[x1][my]==0)
            buffer[x1][my]=(buffer[x1][y1]+buffer[x1][y2])/2
                + random.nextInt()%(variation[depth]/2) - (variation[depth]/4);
        if(buffer[x2][my]==0)
            buffer[x2][my]=(buffer[x2][y1]+buffer[x2][y2])/2
                + random.nextInt()%(variation[depth]/2) - (variation[depth]/4);
        if(buffer[mx][y1]==0)
            buffer[mx][y1]=(buffer[x1][y1]+buffer[x2][y1])/2
                + random.nextInt()%(variation[depth]/2) - (variation[depth]/4);
        if(buffer[mx][y2]==0)
            buffer[mx][y2]=(buffer[x1][y2]+buffer[x2][y2])/2
                + random.nextInt()%(variation[depth]/2) - (variation[depth]/4);
        if(buffer[mx][my]==0)
            buffer[mx][my]=(buffer[x1][y1]+buffer[x2][y1]+buffer[x1][y2]+buffer[x2][y2])/4
                + random.nextInt()%variation[depth] - (variation[depth]/2);

        tessalate(buffer, x1, y1, mx, my, depth-1, variation);
        tessalate(buffer, mx, y1, x2, my, depth-1, variation);
        tessalate(buffer, x1, my, mx, y2, depth-1, variation);
        tessalate(buffer, mx, my, x2, y2, depth-1, variation);
    }

    BranchGroup makeTerrain(double[][] buffer) {
        QuadArray qa=new QuadArray(buffer.length*buffer[0].length*4,
                                   GeometryArray.COORDINATES | GeometryArray.NORMALS);
        //qa.setCapability(Geometry.ALLOW_INTERSECT);
        Vector3f tl=new Vector3f(), tr=new Vector3f(),
            bl=new Vector3f(), br=new Vector3f(),
            x1=new Vector3f(), x2=new Vector3f(),
            normal=new Vector3f();

        for(int y=1; y<buffer[0].length-2; y++) {
            for(int x=1; x<buffer.length-2; x++) {
                /*
                  qa.setCoordinate(y*buffer.length*4 + x*4, new Point3d(x, buffer[x][y], y));
                  qa.setCoordinate(y*buffer.length*4 + x*4 + 1, new Point3d(x+1, buffer[x+1][y], y));
                  qa.setCoordinate(y*buffer.length*4 + x*4 + 2, new Point3d(x+1, buffer[x+1][y+1], y+1));
                  qa.setCoordinate(y*buffer.length*4 + x*4 + 3, new Point3d(x, buffer[x][y+1], y+1));
                */
                qa.setCoordinate(y*buffer.length*4 + x*4, new Point3d(x, buffer[x][y+1], y+1));
                tl.x=x-1; tl.y=(float)buffer[x-1][y];   tl.z=y;
                tr.x=x+1; tr.y=(float)buffer[x+1][y];   tr.z=y;
                bl.x=x-1; bl.y=(float)buffer[x-1][y+2]; bl.z=y+2;
                br.x=x+1; br.y=(float)buffer[x+1][y+2]; br.z=y+2;
                x1.x=tl.x - br.x; x1.y=tl.y - br.y; x1.z=tl.z - br.z;
                x2.x=tr.x - bl.x; x2.y=tr.y - bl.y; x2.z=tr.z - bl.z;
                normal.cross(x2, x1);
                qa.setNormal(y*buffer.length*4 + x*4, normal);

                qa.setCoordinate(y*buffer.length*4 + x*4 + 1, new Point3d(x+1, buffer[x+1][y+1], y+1));
                tl.x=x;   tl.y=(float)buffer[x][y];     tl.z=y;
                tr.x=x+2; tr.y=(float)buffer[x+2][y];   tr.z=y;
                bl.x=x;   bl.y=(float)buffer[x][y+2];   bl.z=y+2;
                br.x=x+2; br.y=(float)buffer[x+2][y+2]; br.z=y+2;
                x1.x=tl.x - br.x; x1.y=tl.y - br.y; x1.z=tl.z - br.z;
                x2.x=tr.x - bl.x; x2.y=tr.y - bl.y; x2.z=tr.z - bl.z;
                normal.cross(x2, x1);
                qa.setNormal(y*buffer.length*4 + x*4 + 1, normal);

                qa.setCoordinate(y*buffer.length*4 + x*4 + 2, new Point3d(x+1, buffer[x+1][y], y));
                tl.x=x;   tl.y=(float)buffer[x][y-1];   tl.z=y-1;
                tr.x=x+2; tr.y=(float)buffer[x+2][y-1]; tr.z=y-1;
                bl.x=x;   bl.y=(float)buffer[x][y+1];   bl.z=y+1;
                br.x=x+2; br.y=(float)buffer[x+2][y+1]; br.z=y+1;
                x1.x=tl.x - br.x; x1.y=tl.y - br.y; x1.z=tl.z - br.z;
                x2.x=tr.x - bl.x; x2.y=tr.y - bl.y; x2.z=tr.z - bl.z;
                normal.cross(x2, x1);
                qa.setNormal(y*buffer.length*4 + x*4 + 2, normal);

                qa.setCoordinate(y*buffer.length*4 + x*4 + 3, new Point3d(x, buffer[x][y], y));
                tl.x=x-1; tl.y=(float)buffer[x-1][y-1]; tl.z=y-1;
                tr.x=x+1; tr.y=(float)buffer[x+1][y-1]; tr.z=y-1;
                bl.x=x-1; bl.y=(float)buffer[x-1][y+1]; bl.z=y+1;
                br.x=x+1; br.y=(float)buffer[x+1][y+1]; br.z=y+1;
                x1.x=tl.x - br.x; x1.y=tl.y - br.y; x1.z=tl.z - br.z;
                x2.x=tr.x - bl.x; x2.y=tr.y - bl.y; x2.z=tr.z - bl.z;
                normal.cross(x2, x1);
                qa.setNormal(y*buffer.length*4 + x*4 + 3, normal);
            }
        }
        TransformGroup tg=new TransformGroup();
        Transform3D t3d=new Transform3D();
        t3d.setScale(new Vector3d(4, .015, 4));
        tg.setTransform(t3d);
        Appearance ap=new Appearance();
        Material mat=new Material(new Color3f((float).1, (float).1, (float).1),
                                  new Color3f((float).0, (float).0, (float).0),
                                  new Color3f((float).8, (float).8, (float).8),
                                  new Color3f((float).4, (float).4, (float).4),
                                  32);
        ap.setMaterial(mat);
        Shape3D land=new Shape3D(qa, ap);
        land.setCollidable(true);
        land.setCapability(Node.ENABLE_COLLISION_REPORTING);
        tg.addChild(land);
        tg.setCapability(Node.ENABLE_COLLISION_REPORTING);
        tg.setCollidable(true);
        BranchGroup bg=new BranchGroup();
        bg.addChild(tg);
        bg.setCapability(Node.ENABLE_COLLISION_REPORTING);
        return bg;
    }

    double Xangle=0.0, Yangle=0.0, Zangle=0.0;

    public void mouseMoved(MouseEvent e) { }
    public void mouseClicked(MouseEvent e) {
        if((e.getModifiers() & InputEvent.BUTTON3_MASK) != 0) {
            Transform3D t3d=new Transform3D();
            t3d.setTranslation(new Vector3d(-10, -23, -10));
            cameraPos.setTransform(t3d);
            mouse3down=true;
        }
    }
    public void mouseEntered(MouseEvent e) { }
    public void mouseExited(MouseEvent e) { }
    public void mouseDragged(MouseEvent e) {
        if((e.getModifiers() & InputEvent.BUTTON1_MASK) != 0) {
            Transform3D crotx=new Transform3D();
            Transform3D croty=new Transform3D();
            Transform3D crotz=new Transform3D();

            double deltaX=(oldMouseX-e.getX());
            double deltaY=(oldMouseY-e.getY());

            Yangle-=(deltaX*.01);
            Xangle-=(deltaY*.01);

            crotx.rotX(Xangle);
            croty.rotY(Yangle);
            crotz.rotZ(Zangle);
            crotx.mul(croty);
            crotx.mul(crotz);
            cameraRot.setTransform(crotx);

            oldMouseX=e.getX();
            oldMouseY=e.getY();
        }
    }
    public void mousePressed(MouseEvent e) {
        if((e.getModifiers() & InputEvent.BUTTON1_MASK) != 0) {
            mouse1down=true;
            movement.setEnable(true);
            oldMouseX=e.getX();
            oldMouseY=e.getY();
        }
        if((e.getModifiers() & InputEvent.BUTTON2_MASK) != 0) {
            if(!movement.isMoving()) {
                mouse2down=true;
                movement.startMovement(new Vector3d(0.0, 0.0, .5), 50);
                movement.setEnable(true);
                movement.setMoving(true);
            }
        }

        if((e.getModifiers() & InputEvent.BUTTON3_MASK) != 0) {
            mouse3down=true;
        }
    }
    public void mouseReleased(MouseEvent e) {
        if((e.getModifiers() & InputEvent.BUTTON1_MASK) != 0) {
            mouse1down=false;
        }
        if((e.getModifiers() & InputEvent.BUTTON2_MASK) != 0) {
            mouse2down=false;
            movement.setMoving(false);
        }
        if((e.getModifiers() & InputEvent.BUTTON3_MASK) != 0) {
            mouse3down=false;
        }

        if(!(mouse1down || mouse2down)) movement.setEnable(false);
    }


    public static void main(String[] args) {
        new Terrain();
    }
}
